<?php

/*
Copyright 2013 Bing Technologies Pty Lid. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE BING TECHNOLOGIES PTY.LTD.  "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL BING TECHNOLOGIES PTY. LTD. BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

The views and conclusions contained in the software and documentation are
those of the authors and should not be interpreted as representing official
policies, either expressed or implied, of Bing Technologies Pty. Ltd.
*/

class ep_uploader {

	// session control endpoints
	const SESS_ID_NAME = "publicinterface/get_session_id";
	const SESS_END_NAME = "publicinterface/end_session";
	// upload endpoint prefix
	const UPLOAD_PREFIX = "direct_upload";

	private $EASYPOST_HOSTS = array('https://easymail.com', 'https://alt.easymail.com');

	//configuration array
	private $conf = array(
		"AUTH_USER" => NULL,
		"AUTH_PASS" => NULL,
		"ACCOUNT" => NULL,
		"PROXY" => NULL,
		"VERIFY_CERTIFICATE" => true
	);

	public function version() {
		return '1.3';
	}

	// to load configuration from an ini like config file
	public function load_config_ini($path) {
		$conf = parse_ini_file($path);
		foreach( array_keys($this->conf) as $ck ) {
			if (in_array($ck, array_keys($conf))) {
				$this->conf[$ck] = $conf[$ck];
			}
		}
	}

	// to load configuration from the specific php file
	public function load_config_php($path) {
		require($path);
		$class = new ReflectionClass("ep_api_conf");
		foreach( array_keys($this->conf	) as $ck ) {
			if ($class->hasConstant(strtoupper($ck))) {
				$this->conf[$ck] = $class->getConstant(strtoupper($ck));
			}
		}
	}

	// to set or unset echoing of debug information during the library work
	public function allow_debug($allow=true)
	{
		$this->DEBUG = $allow;
	}

	// to set the debug echo lines separator
	// for shell the default \n is good
	// for web it is better to use <br/>
	public function set_debug_echo_line_separator($separator='<br/>')
	{
		$this->echo_separator = $separator;
	}

	// to create json temporary file for a data array with adding some fields from the config file if they are not already set in $data_array
	public function make_json_file($data_array)
	{
		$rs = tempnam(sys_get_temp_dir(), 'easypost_');
		$fh = fopen($rs, "w");
		fwrite($fh, json_encode($data_array));
		fclose($fh);
		return $rs;
	}

	// the main function for the upload
	// $opts:
	// http_hosts - one server address as a string or an array of server addresses to override the default
	// account - account to override the default from the config file
	// auth_user - user to override the default from the config file
	// auth_pass - password to override the default from the config file
	// chunk_size - size of a peace of a file that is transfered in one go
	public function do_upload($fnames, $opts=array())
	{
		// remove old server digest authentication data
		unset($this->server_digest);
		if ( ! is_array($fnames) ) {
			$fnames = array($fnames);
		}
		// count meta file names
		$meta_count = 0;
		for($i=0; $i<count($fnames); $i++) {
			if( ! is_array($fnames[$i]) ) {
				$fnames[$i] = array($fnames[$i], basename($fnames[$i]));
			}
			$fnames[$i][1] = strtr(basename($fnames[$i][1]), " ^*!$\"'`(),<>[]\\|{}", "___________________");
			//check that no xml, text and json files more than 1 in $fnames
			$ext = pathinfo($fnames[$i][1], PATHINFO_EXTENSION);
			if( in_array($ext, $this->metas) ) {
				$meta_count++;
				if($meta_count > 1) {
					throw new Exception('No more than 1 meta-data file allowed'."\n");
				}
			}
		}
		// make options using configuration file data overwriting it by the function parameter data
		$opts = array_replace(array('chunk_size' => 65535, 'auth_user' => $this->conf['AUTH_USER'], 'auth_pass' => $this->conf['AUTH_PASS'], 'account' => $this->conf['ACCOUNT']), $opts);
		// temporary commented out defaults for hosts
		$opts = array_replace(array('chunk_size' => 65535, 'http_hosts' => $this->EASYPOST_HOSTS), $opts);
		if( is_string($opts['http_hosts']) ) {
			$opts['http_hosts'] = array($opts['http_hosts']);
		}
		$http_host = $this->find_the_host($opts['http_hosts']);
		// get session id
		$reply = json_decode($this->extract_body($this->get($http_host, '/'.$this::SESS_ID_NAME.'.json?account='.$opts['account'], $opts['auth_user'], $opts['auth_pass'])), true);
		$session_id = $reply['session_id'];
		// upload files
		foreach( $fnames as $afname ) {
			try {
				$this->put($http_host, '/'.$this::UPLOAD_PREFIX.'/'.$session_id.'/'.rawurlencode($afname[1]), $opts['auth_user'], $opts['auth_pass'], $afname[0], $opts['chunk_size']);
			} catch(Exception $e) {
				$this->echo_log($e);
			}
		}
		// close the session
		return json_decode($this->extract_body($this->get($http_host, '/'.$this::SESS_END_NAME.'.json?session_id='.$session_id, $opts['auth_user'], $opts['auth_pass'])), true);
	}

	// echo function that respects the debug mode
	public function echo_log($string)
	{
		if ( $this->DEBUG ) {
			if ( $this->echo_separator ) {
				$string = str_replace("\n", $this->echo_separator, $string);
			}
			echo $string;
		}
	}

//################################## private stuff lower #######################################################

	// on/off debug mode
	private $DEBUG = false;
	// debug mode echo line separator (it's good to have it <br/> in case of debugging web forms)
	private $echo_separator = "\n";
	// storage for WWW-Authenticate header value of the server
	private $server_digest;
	// list of meta file extensions
	private $metas = array('xml', 'json', 'text', 'txt');

	// returns file size, if the file does not exist it throws an exception
	private function filesize($fname)
	{
		if(!file_exists($fname)) {
			throw new Exception('File "'.$fname.'" not found'."\n", 404);
		}
		return filesize($fname);
	}

	// find and return available host in the list
	private function find_the_host($host_list)
	{
		foreach($host_list as $http_host) {
			try {
				$this->head($http_host);
			} catch(Exception $e) {
				$this->echo_log($e);
				$http_host = '';
			}
		}
		if ( empty($http_host) ) {
			throw new Exception('No connectable host found'."\n");
		}
		return $http_host;
	}

	// extract from response and store the server digest header
	private function store_server_digest($response)
	{
		$headers_in = $this->extract_headers($response);
		$digest = explode(' ', $headers_in['WWW-Authenticate']);
		if ($digest[0] == 'Digest') {
			$this->server_digest = array('nc' => 0, 'digest' => $digest);
		} else {
			throw new Exception('Digest authentication response header expected'."\n");
		}
	}

	// make digest header line using header data from the web server response
	private function make_digest_header($method, $uripath, $auth_user, $auth_pass)
	{
		if(isset($this->server_digest)) {
			$digest = $this->server_digest['digest'];
			$nc = ++$this->server_digest['nc'];
			$cnonce = uniqid();
			$dparams = array();
			foreach ( array_slice($digest, 1) as $el) {
				$el = explode('=', rtrim($el, ','));
				$dparams[$el[0]] = trim($el[1], '"');
			}
			$ha1 = md5($auth_user.':'.$dparams['realm'].':'.$auth_pass);
			$ha2 = md5("$method:$uripath");
			$response_ha = $ha1.':'.$dparams['nonce'].':'.sprintf('%08.8d', $nc).':'.$cnonce.':'.$dparams['qop'].':'.$ha2;
			$response_ha = md5($response_ha);
			$digest_h = 'Digest username="'.$auth_user.'", realm="'.$dparams['realm'].'", nonce="'.$dparams['nonce'].'", uri="'.$uripath.'", response="'.$response_ha.'", qop='.$dparams['qop'].', nc='.sprintf('%08.8d', $nc).', cnonce="'.$cnonce.'"';
			$rs = 'Authorization: '.$digest_h;
		} else {
			$rs = '';
		}
		return $rs;
	}

	//to create an assosiative array from a simple header array
	private function array2array($array_in)
	{
		$rs = array();
		foreach ($array_in as $ar) {
			$arr = explode(':', $ar);
			if (count($arr) == 1) {
				array_push($rs, trim($ar));
			} else {
				$rs[$arr[0]] = trim(implode(':', array_slice($arr, 1)));
			}
		}
		return $rs;
	}

	//to extract headers part from an http response
	private function extract_headers($response)
	{
		$rs = explode("\r\n\r\n", $response);
		return $this->array2array(explode("\n", $rs[0]));
	}

	//to get body part from an http response
	private function extract_body($response)
	{
		$rs = explode("\r\n\r\n", $response);
		return $rs[1];
	}

	//to get the first line an http response that contains the http code
	private function get_response_first_line($headers_in)
	{
		$l = explode("\n", $headers_in);
		return $l[0];
	}

	//to get the http code from a response
	private function get_response_code($headers_in)
	{
		preg_match('|HTTP/\d\.\d\s+(\d+)\s+.*|', $headers_in, $match);
		return $match[1];
	}

	//create proper array for the server url filling gaps with defaults
	private function parse_server_url($_server)
	{
		$rs = parse_url($_server);
		if (!array_key_exists('scheme', $rs)) {
			$rs['scheme'] = 'http';
		}
		if (!array_key_exists('port', $rs)) {
			if ( $rs['scheme'] == 'https' ) {
				$rs['port'] = 443;
			} else {
				$rs['port'] = 80;
			}
		}
		$rs['socket_host'] = ( $rs['scheme'] == 'https' ) ? 'ssl://'.$rs['host'] : $rs['host'];
		return $rs;
	}

	//GET http method function
	private function get($_server, $uri, $user, $pass)
	{
		$digest_header = $this->make_digest_header('GET', $uri, $user, $pass);
		$fp = $this->fsockopen($_server, $errno, $errstr, 30);
		if ($fp)
		{
			$p = $this->method_line('GET', $_server, $uri, '1.0');
			// it always expects that the connection is digest authenticated
			if ( $digest_header ) {
				$p .= $digest_header."\r\n";
			}
			$sdata = $this->parse_server_url($_server);
			$p.= "Host: ".$sdata['host'].':'.$sdata['port']."\r\n";
			$p.= "Content-length: 0\r\n";
			$p.= "\r\n";
			$this->echo_log("\nRequest headers:\n".$p);
			fputs ($fp, $p);
		} else {
			throw new Exception($uri.":".(($errno == 0) ? 'Unknown error' : $errstr)."\n", $errno);
		}
		// get response
		$rs = "";
		while ($l=fgets($fp)) {
			$rs .= $l;
		}
		$this->echo_log("\nResponse:\n".$rs."\n\n");
		fclose($fp);
		if( $rs == '' ) {
			throw new Exception($uri.":Empty response\n");
		}
		// extract response code
		$code = $this->get_response_code($rs);
		if ( $code == '401' ) {
			if ( $digest_header == '' ) {
				$this->store_server_digest($rs);
				$rs = $this->get($_server, $uri, $user, $pass);
			} else {
				throw new Exception($uri.":".$this->get_response_first_line($rs)."\n", $code);
			}
		} elseif ( $code != '200' ) {
			// an error has happened
			throw new Exception($uri.":".$this->get_response_first_line($rs)."\n", $code);
		}
		return $rs;
	}

	// PUT http method function
	// In case of Digest Authentication a GET method should be called first so the server digest parameters have already be stored
	private function put($_server, $uri, $user, $pass, $fname, $chunk_size=65535)
	{
		$fp = $this->fsockopen($_server, $errno, $errstr, 30);
		if ($fp)
		{
			// it always expects that the connection is digest authenticated
			$digest_header = $this->make_digest_header('PUT', $uri, $user, $pass);
			$fsize = $this->filesize($fname);
			$sdata = $this->parse_server_url($_server);
			$p = $this->method_line('PUT', $_server, $uri, '1.1');
			if ( $digest_header ) {
				$p .= $digest_header."\r\n";
			}
			$p.= "Host: ".$sdata['host'].':'.$sdata['port']."\r\n";
			$p.= "Connection: close\r\n";
			$p.= "Content-length: ".$fsize."\r\n";
//Expect does not work for forward http proxy
//			$p.= "Expect: 100-continue\r\n";
			$p.= "\r\n";

			$this->echo_log("\nRequest headers:\n".$p);
			fputs ($fp, $p);
		} else {
			throw new Exception($uri.":\n".(($errno == 0) ? 'Unknown error' : $errstr)."\n");
		}
		// send file
		$fh = fopen($fname, 'rb');
		$this->echo_log("Sending $fsize bytes of the file $fname ...\n");
		$tot = 0;
		while(!feof($fh)) {
			$data = fread($fh, $chunk_size);
			$tot += strlen($data);
			$this->echo_log($tot." of ".$fsize." bytes of $fname ... ");
			fwrite($fp, $data);
			$this->echo_log("sent\n");
		}
		fclose($fh);
		// get response
		$rs = "";
		while ($l=fgets($fp)) {
			if(trim($l)) {
				$rs .= $l;
			}
		}
		fclose($fp);
		$this->echo_log("\nResponse:\n".$rs."\n\n");
		// extract the return code
		$code = $this->get_response_code($rs);
		if ( $code != '201' ) {
			// an error has happened
			throw new Exception($uri.":\n".$this->get_response_first_line($rs)."\n", $code);
		}
		return $rs;
	}

	// HEAD http method function to test ability to communicate with a host
	private function head($_server)
	{
		$fp = $this->fsockopen($_server, $errno, $errstr, 30);
		if ($fp)
		{
			$sdata = $this->parse_server_url($_server);
			$p = $this->method_line('HEAD', $_server, '/', '1.0');
			$p .= "Host: ".$sdata['host'].':'.$sdata['port']."\r\n";
			$p .= "\r\n";
			$this->echo_log("\nRequest headers:\n".$p);
			fputs ($fp, $p);
		}
		else throw new Exception($uri.":\n".$errstr."\n");

		$rs = "";
		while ($l=fgets($fp)) {
			$rs .= $l;
		}
		$this->echo_log("\nResponse:\n".$rs."\n\n");
		fclose($fp);
		return $rs;
	}

	// begin request differently depending on whether proxy is set or not
	private function method_line($method, $_server, $uri, $version)
	{
		$rs = $method.' '.$uri.' HTTP/'.$version."\r\n";
		if ( $this->conf['PROXY'] ) {
			$sdata = $this->parse_server_url($_server);
			if ( $sdata['scheme'] != 'https' ) {
				$rs = $method.' '.$_server.$uri.' HTTP/'.$version."\r\n";
			}
		}
		return $rs;
	}

	// open and configure socket depending on http or https protocols
	// and whether proxy is set or not
	private function fsockopen($_server, &$errno, &$errstr, $timeout)
	{
		// if the proxy is set then use the proxy host for socket
		if ( $this->conf['PROXY'] ) {
			$sdata = $this->parse_server_url($this->conf['PROXY']);
		} else {
			$sdata = $this->parse_server_url($_server);
		}
		$context = stream_context_create();
		// set verification of the server certificate if it is configured
		if ( $this->conf['VERIFY_CERTIFICATE'] ) {
			stream_context_set_option($context, 'ssl', 'verify_peer', true);
		}
		// create actual socket
		$rs = stream_socket_client($sdata['socket_host'].':'.$sdata['port'], $errno, $errstr, $timeout, STREAM_CLIENT_CONNECT, $context);
		// if the proxy is set ...
		if ( $this->conf['PROXY'] ) {
			$sdata = $this->parse_server_url($_server);
			// and protocol is https then use CONNECT command
			if ( $sdata['scheme'] == 'https' ) {
				$p = 'CONNECT '.$sdata['host'].':'.$sdata['port']." HTTP/1.0\r\n";
				$p .= "User-Agent: Wget/1.12 (freebsd7.3)\r\n";
				$p .= "Proxy-Connection: Keep-Alive\r\n\r\n";
				$this->echo_log("\nRequest headers:\n".$p);
				fputs ($rs, $p);
				$r = "";
				while ($l=fgets($rs)) {
					if(trim($l)) {
						$r .= $l;
					} else {
						break;
					}
				}
				$this->echo_log("\nProxy reply:\n".$r);
				// after proxy has establish a connection to the real host
				// the socket should become ssl encrypted
				stream_socket_enable_crypto($rs, TRUE, STREAM_CRYPTO_METHOD_SSLv23_CLIENT);
			}
		}
		return $rs;
	}

};

?>
